#ifndef DOM_PARALLEL_DISTMEM_HPP
#  define DOM_PARALLEL_DISTMEM_HPP

#  include "dom_parallel.hpp"

template <class unit, typename real_t, class shrdmem_class>
class dom_parallel_distmem : public shrdmem_class
{
  // <nested class>
  private: class dom_halo : public dom<unit, real_t>
  {
    private: dom_parallel_distmem<unit, real_t, shrdmem_class> *nghbr;
    private: quantity<unit, real_t> *ledge, *redge;
    private: bool valid;
    private: int halo, left, rght;

    public: dom_halo(int halo, dom_parallel_distmem<unit, real_t, shrdmem_class> *nghbr)
      : nghbr(nghbr), valid(false), halo(halo)
    {
      ledge = new quantity<unit, real_t>[halo];
      redge = new quantity<unit, real_t>[halo];
      left = (nghbr->size + nghbr->rank - 1) % nghbr->size, 
      rght = (nghbr->size + nghbr->rank + 1) % nghbr->size;
    }

    public: ~dom_halo() 
    {
      delete[] ledge;
      delete[] redge;
    }

    public: virtual quantity<unit, real_t> *left_edge(int n) { xchnge_halos(n); return ledge; }
    public: virtual quantity<unit, real_t> *rght_edge(int n) { xchnge_halos(n); return redge; }
    public: virtual void integ_loop(int, quantity<si::dimensionless, real_t>&) { assert(false); }

    public: void xchnge_halos(int n)
    {
      if (!valid) nghbr->distmem_xchnge_halos(left, rght, halo, n, ledge, redge);
      valid = !valid;
    }
  };
  // </nested class>

  private: int size, rank;
  private: dom_halo *halodomain;

  public: dom_parallel_distmem(adv<unit, real_t> *fllbck, adv<unit, real_t> *advsch, 
    int i_min, int i_max, int size, int rank)
    : shrdmem_class(fllbck, advsch, 
      i_min +    rank    * (i_max - i_min + 1) / size, 
      i_min + (rank + 1) * (i_max - i_min + 1) / size - 1,
      1 // FIXME: MPI+OpenMP and MPI+threads 
    ), size(size), rank(rank)
  {
    // subdomain length
    int nx = i_max - i_min + 1, nxs = nx / size; 
    if (nxs != ((1.*nx) / (1.*size))) error_macro("nx/nk must be an integer value (" << nx << "/" << size << " given)")

    halodomain = new dom_halo((advsch->stencil_extent() - 1) / 2, this);
    this->hook_neighbours(halodomain, halodomain);
  }

  public: ~dom_parallel_distmem()
  {
    delete halodomain;
  }

  private: virtual void distmem_xchnge_halos(int left, int rght, int halo, int n,
    quantity<unit, real_t> *lhalo, quantity<unit, real_t> *rhalo) = 0;
  protected: virtual void distmem_string_snd(int k, string) = 0;
  protected: virtual void distmem_string_rcv(int k, string&) = 0;
  private: virtual void distmem_barrier() = 0;

  private: void barrier() 
  {
    shrdmem_class::barrier();
    distmem_barrier();
  }

  public: void output(ostream &sout, int n, bool endrec = true) 
  {
    if (rank == 0) shrdmem_class::output(sout, n, false);
    for (int k=1; k < size; ++k)
    {   
      distmem_barrier();
      if (k == rank)
      {   
        ostringstream tmp;
        shrdmem_class::output(tmp, n, false);
        distmem_string_snd(0, tmp.str());
      }   
      else if (rank == 0)
      {   
        string tmp;
        distmem_string_rcv(k, tmp);
        sout << tmp;
      }   
    }   
    distmem_barrier();
    if (rank == 0 && endrec) sout << endl << endl;
  }
};
#endif
