#ifndef DOM_PARALLEL_DISTMEM_MPI_HPP
#  define DOM_PARALLEL_DISTMEM_MPI_HPP

#  include <boost/serialization/string.hpp> // serialization of strings
#  include <boost/units/io.hpp> // serialization of units
#  include <boost/mpi.hpp>
namespace mpi = boost::mpi;

#  include "dom_parallel_distmem.hpp"

template <class unit, typename real_t, class shrdmem_class>
class dom_parallel_distmem_mpi : public dom_parallel_distmem<unit, real_t, shrdmem_class>
{
  private: enum msgtype { msg_halo_l, msg_halo_r, msg_out };
  private: mpi::communicator *comm;
  private: mpi::environment *env;

  /// MPI init for using rank as an argument to the parent class c-tor
  private: unsigned mpi_init()
  {
    env = new mpi::environment(); // FIXME: use MPI::Init_thread(MPI_THREAD_MULTIPLE) instead to support MPI+threads?
    comm = new mpi::communicator();
    return comm->rank();
  }

  public: dom_parallel_distmem_mpi(adv<unit, real_t> *fllbck, adv<unit, real_t> *advsch, int i_min, int i_max, int nk)
    : dom_parallel_distmem<unit, real_t, shrdmem_class>(fllbck, advsch, i_min, i_max, nk, mpi_init())
  {
    if (nk != comm->size()) 
      error_macro("MPI is using different number of nodes (" << comm->size() << ") than specified (" << nk << ")")
  }

  public: ~dom_parallel_distmem_mpi()
  {
    delete comm;
    delete env; // FIXME: use MPI::Finalize() instead to support MPI+threads?
  }

  private: void distmem_barrier() 
  {
    comm->barrier();
  }

  private: void distmem_string_snd(int k, string msg)
  {
    comm->send(k, msg_out, msg);
  }

  private: void distmem_string_rcv(int k, string &msg) 
  {
    comm->recv(k, msg_out, msg);
  }

  private: void distmem_xchnge_halos(int left, int rght, int halo, int n,
    quantity<unit, real_t> *ledge, quantity<unit, real_t> *redge) 
  {
    mpi::request reqs[4];
    int r = 0;
    // isend() / irecv() are non-blocking
    reqs[r++] = comm->isend(left, msg_halo_l, this->left_edge(n), halo);
    reqs[r++] = comm->irecv(rght, msg_halo_l, ledge, halo);
    reqs[r++] = comm->isend(rght, msg_halo_r, this->rght_edge(n), halo);
    reqs[r++] = comm->irecv(left, msg_halo_r, redge, halo);
    mpi::wait_all(reqs, reqs + r);
  }
};
#endif
