#ifndef DOM_PARALLEL_OPENMP_HPP
#  define DOM_PARALLEL_OPENMP_HPP

#  include "dom_parallel.hpp"
#  include <omp.h>

template <class unit, typename real_t>
class dom_parallel_openmp : public dom_parallel<unit, real_t>
{
  public: dom_parallel_openmp(adv<unit, real_t> *fllbck, adv<unit, real_t> *advsch, int i_min, int i_max, int nk)
    : dom_parallel<unit, real_t>(fllbck, advsch, i_min, i_max, nk)
  {
    int ncpu = omp_get_num_procs();
    if (nk > ncpu) warning_macro("using more threads (" << nk << ") than CPUs/cores (" << ncpu << ")")
    omp_set_num_threads(nk);
  }

  public: void barrier()
  {
#  pragma omp barrier
  }

  public: void integ_loop(int nt, quantity<si::dimensionless, real_t> &C)
  {
    int k; 
#  pragma omp parallel private(k)
    {   
      k = omp_get_thread_num();
      this->integ_loop_k(nt, C, k); 
    } 
  }
  
};
#endif
