#ifndef DOM_PARALLEL_THREADS_HPP
#  define DOM_PARALLEL_THREADS_HPP

#  include "dom_parallel.hpp"
#  include <boost/thread.hpp>

template <class unit, typename real_t>
class dom_parallel_threads : public dom_parallel<unit, real_t>
{
  private: boost::barrier *b; 
  private: int nk;

  public: dom_parallel_threads(adv<unit, real_t> *fllbck, adv<unit, real_t> *advsch, int i_min, int i_max, int nk)
    : dom_parallel<unit, real_t>(fllbck, advsch, i_min, i_max, nk), nk(nk)
  {
    int ncpu = boost::thread::hardware_concurrency();
    if (nk > ncpu) warning_macro("using more threads (" << nk << ") than CPUs/cores (" << ncpu << ")")
    b = new boost::barrier(nk);
  }

  public: ~dom_parallel_threads() { delete b; }

  public: void barrier() { b->wait(); }

  public: void integ_loop(int nt, quantity<si::dimensionless, real_t> &C)
  {
    boost::thread_group threads;
    for (int k = 0; k < nk; ++k) threads.add_thread(new boost::thread(
      &dom_parallel<unit, real_t>::integ_loop_k, this, nt, C, k
    ));
    threads.join_all();
  }
};
#endif
